#include "squidwall.h"

/* filteroptions for content scanning
 *
 * 1. content filtering
 * +C,Filter1,Filter2,Filter3 (regex content filtering)
 * -C disable content filtering
 *
 * 2. url filtering
 * +U,Filter1,Filter2,Filter3 (regex url filtering)
 * -U disable url filtering
 *
 * 3. antivirus scanning
 * +V
 * -V disable antivirus scanning
 *
 * DEFAULT:
 * +C,*
 * +U,*
 * +V
 */

static void read_regex_freefilter(array *root);
static void read_regex_allfiles(array *a);
static void read_regex_file(array *root, char *file);
static void read_regex(array *a, char *start);

static void read_regex_freefilter(array *root)
{
	struct squidwall_regex *r;
	int64 l;

	for (;;) {
		l=array_length(root, sizeof(struct squidwall_regex));
		if (l == 0) break;

		r = array_get(root, sizeof(struct squidwall_regex), l-1);
		if (!r) {
			panic("got an empty array from array_get()!");
			break;
		}

		/* free allocated things */
		stralloc_free(&r->name);
		regfree(&r->r);

		/* destroy unsued array members */
		array_truncate(root, sizeof(struct squidwall_regex), l-1);
	}
	/* free tho whole array */
	array_reset(root);

	return;
}

static void read_regex_allfiles(array *a)
{
	DIR *D;
	struct dirent *d;

	D=opendir(filters_dir);
	if (!D) return;
	while ((d=readdir(D))) {

		/* skip '.' and '..' */
		if (d->d_name[0]=='.' && d->d_name[1]==0) continue;
		if (d->d_name[0]=='.' && d->d_name[1]=='.' && d->d_name[2]==0)
			continue;

		read_regex_file(a, d->d_name);
	}
	closedir(D);

	return;
}

static void read_regex_file(array *root, char *file)
{
	struct squidwall_regex *r;
	stralloc sa;
	buffer b;

//	debug_s(9, "read_regex_file() file=", file);
	stralloc_init(&sa);

	if (!stralloc_copys(&sa, filters_dir)) die_nomem();
	if (!stralloc_cats(&sa, "/")) die_nomem();
	if (!stralloc_cats(&sa, file)) die_nomem();
	sa.s[sa.len]=0;

	/* read /etc/squidwall/hosts/+192.168.100.5 e.g. */
	if (buffer_mmapread(&b, sa.s) == -1) {
		log_headline("error");
		log_s("filter '"); log_sa(&sa); log_s("' not found or empty!\n");
		stralloc_free(&sa); return;
	}

	while (buffer_getnewline_sa(&b, &sa) == 1) {
		sa.s[sa.len-1]=0; /* remove newline */
		/* ... add to regex filter */
		r=array_allocate(root, sizeof(struct squidwall_regex),
			array_length(root, sizeof(struct squidwall_regex)));
		if (!r) die_nomem();
		byte_zero(r, sizeof(struct squidwall_regex));
		if (regcomp(&r->r, sa.s, SW_REGCOMP_CFLAGS) != 0) {
			panic("regcomp() failed!");
			continue; /* read next line */
		}

		/* save the filtername (is used for the error reporting website) */
		stralloc_init(&r->name);
		if (!stralloc_copys(&r->name, file)) die_nomem();
		if (!stralloc_cats(&r->name, " (")) die_nomem();
		if (!stralloc_cats(&r->name, sa.s)) die_nomem();
		if (!stralloc_cats(&r->name, ")")) die_nomem();
		r->name.s[r->name.len]=0;
	}
	buffer_close(&b);
	stralloc_free(&sa);

	return;
}

static void read_regex(array *a, char *start)
{
	char *x=start;
	int i=0, l=str_len(start);

	while (*++x) {

		/* read all available filters and return */
		if (*x == '*') {
			read_regex_allfiles(a);
			return;
		}

		/* ',' -> '\0' */
		if (*x == ',') *x = 0;
	}

	/* read and compile each given filter */
	for (x=start+1; i < l; i+=str_len(x+i)+1) {
		read_regex_file(a, x+i);
	}

	return;
}

void read_filters(struct squidwall_acl *acl, const char *path, const char *file)
{
	stralloc sa;
	buffer b;

	stralloc_init(&sa);
	if (!stralloc_copys(&sa, path)) die_nomem();
	if (!stralloc_cats(&sa, "/")) die_nomem();
	if (!stralloc_cats(&sa, file)) die_nomem();
	sa.s[sa.len]=0;

	/* read /etc/squidwall/hosts/+192.168.100.5 e.g. */
	if (buffer_mmapread(&b, sa.s) == -1) {
		/* if the host/user failed... read the '!'
		 * - /etc/squidwall/hosts/!
		 * - /etc/squidwall/users/!
		 */
		if (!stralloc_copys(&sa, path)) die_nomem();
		if (!stralloc_cats(&sa, "/!")) die_nomem();
		sa.s[sa.len]=0;
		if (buffer_mmapread(&b, sa.s) == -1) return;
	}

	/* set default values for the acl */
	acl->todo=ACL_DEFAULT;

	for (;;) {
		if (buffer_getnewline_sa(&b, &sa) != 1) break; /* 0=EOF -1=error 1=ok */
		sa.s[sa.len-1]=0; /* remove newline */

		if (sa.len < 2) continue; /* should not happen */
		switch (sa.s[0]) {
		case '+':
			switch (sa.s[1]) {
			case 'C':
				set(acl->todo, TODO_CONT);
				read_regex(&acl->cfilter, sa.s+2);
				break;
			case 'U':
				set(acl->todo, TODO_URL);
				read_regex(&acl->ufilter, sa.s+2);
				break;
			case 'V':
				set(acl->todo, TODO_VIRUS);
				break;
			} /* switch sa.s[1] */
			break;
		case '-':
			switch (sa.s[1]) {
			case 'C':
				unset(acl->todo, TODO_CONT);
				read_regex_freefilter(&acl->cfilter);
				break;
			case 'U':
				unset(acl->todo, TODO_URL);
				read_regex_freefilter(&acl->ufilter);
				break;
			case 'V':
				unset(acl->todo, TODO_VIRUS);
				break;
			} /* switch sa.s[1] */
			break;
		} /* switch sa.s[0] */
	}
	buffer_close(&b);
	stralloc_free(&sa);

	return;
}
